#!/bin/python3
# python 3.8+ only

from Cryptodome.Util import number
from Cryptodome.Hash import SHA512
import random
import math

def rsa_generate_key(n_length: int) -> \
        tuple[tuple[int, int, int], tuple[int, int]]:
    """Return an RSA key pair with primes of length n/2

    The return value is a tuple containing two tuples:
      1. The first tuple is the private key, containing (p, q, d).
      2. The second tuple is the public key, containing (n, e).
    """
    # Compute the product of p and q
    p = number.getPrime(n_length//2)
    q = number.getPrime(n_length//2)
    n = p * q
    phi_n = (p - 1) * (q - 1)

    # Choose e such that gcd(e, phi_n) == 1.
    # Since e is chosen randomly, we repeat the random choice
    # until e is coprime to phi_n.
    e = random.randint(2, phi_n - 1)
    while math.gcd(e, phi_n) != 1:
        e = random.randint(2, phi_n - 1)

    # Choose d such that e * d % phi_n = 1 (modular inverse)
    d = pow (e, -1, phi_n)

    return ((p, q, d), (n, e))


def bighash (enc : bytes, nbits: int, iteration: int) -> int:
    """Hash the encoded inputs into a value with nbits, including
       an iteration counter when hashing to possibly produce
       multiple different hash functions from the same input.
    """
    hx = SHA512.new(enc)
    # Add the iteration counter
    hx.update (str(iteration).encode('utf-8'))
    # Add a separator to clearly separate the iteration counter
    # from the actual payload (the character does not matter).
    hx.update ("%".encode('utf-8'))
    # Build a hash output that is (more than) long enough
    resulta = []
    for i in range(1 + (nbits // 512)):
        # We append i to the hash input in every iteration to randomize the output.
        hx.update (str(i).encode('utf-8'))
        resulta.append (hx.digest ())
    # Combine all of the individual byte arrays
    result = b"".join(resulta)
    # Convert to binary
    binres = format (int.from_bytes (result, "big"), 'b')
    # Truncate to return exactly nbits
    bintrunc = binres[:nbits]
    # Convert back to binary
    return int(bintrunc, 2)


def fdh (message: str, n: int) -> int:
    nbits = len(bin(n)[2:])
    enc = message.encode('utf-8')
    iter = 0
    while True:
        hx = bighash (enc, nbits, iter)
        # The output must be < n
        if (hx < n):
            break
        # We (deterministically) hash again with a
        # fresh 'iter' value to try to get below n
        # while maintaining perfect distribution
        iter = iter + 1
    return hx

def rsa_verify(public_key: tuple[int, int], signature: int, message: str) -> int:
    """Verify the given signature over message using the signer's public_key.

    Preconditions:
        - public_key is a valid RSA public key (n, e)
        - 0 < plaintext < public_key[0]
    """
    n, e = public_key[0], public_key[1]

    pth1 = fdh (message, n)
    pth2 = pow (signature, e, n) # (signature ** e) % n

    return (pth1 == pth2)


def rsa_sign(private_key: tuple[int, int, int], message: str) -> int:
    """Sign the given message using the signer's private key.

    Preconditions:
        - private_key is a valid RSA private key (p, q, d)
    """
    p, q, d = private_key[0], private_key[1], private_key[2]
    n = p * q

    pth = fdh (message, n)
    signature = pow (pth, d, n) # (pth ** d) % n

    return signature

if __name__ == "__main__":
    n_bits = 2048
    message = "hello"

    priv_key, pub_key = rsa_generate_key (n_bits)

    signature = rsa_sign (priv_key, message)

    if (rsa_verify (pub_key, signature, message)):
        print("Signature OK")
    else:
        print("Signature invalid")
